#!/usr/bin/python3
"""
// © University of Southampton IT Innovation Centre, 2018
//
// Copyright in this software belongs to University of Southampton
// IT Innovation Centre of Gamma House, Enterprise Road,
// Chilworth Science Park, Southampton, SO16 7NS, UK.
//
// This software may not be used, sold, licensed, transferred, copied
// or reproduced in whole or in part in any manner or form or in or
// on any media by any person other than in accordance with the terms
// of the Licence Agreement supplied with the software, or otherwise
// without the prior written consent of the copyright owners.
//
// This software is distributed WITHOUT ANY WARRANTY, without even the
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE, except where stated in the Licence Agreement supplied with
// the software.
//
//      Created By :            Nikolay Stanchev
//      Created Date :          09-07-2018
//      Created for Project :   FLAME
"""

from json import dumps, loads
from psutil import NoSuchProcess, STATUS_SLEEPING
from unittest.mock import patch, Mock, MagicMock, PropertyMock
import pytest
from pyramid import testing
from pyramid.httpexceptions import HTTPBadRequest, HTTPNotFound, HTTPInternalServerError
from clmcservice.graphapi.views import GraphAPI
from clmcservice.models import MonitoringProcess
from clmcservice.graphapi.conftest import links, sdn_switches, switches, ues, clusters
from clmcservice.graphapi.utilities import delete_network_graph


class TestGraphAPI(object):
    """
    A pytest-implementation test for the Graph API endpoints.
    """

    @pytest.fixture(autouse=True)
    def app_config(self):
        """
        A fixture to implement setUp/tearDown functionality for all tests by initializing configuration structure for the web service and db connection
        """

        self.registry = testing.setUp()
        self.registry.add_settings({"neo4j_host": "localhost", "neo4j_port": 7687, "neo4j_password": "admin", "influx_host": "localhost", "influx_port": 8086, "network_bandwidth": 104857600,
                                    "network_ues_path": "/opt/clmc/src/service/resources/GraphAPI/network_ues.json",
                                    "network_clusters_path": "/opt/clmc/src/service/resources/GraphAPI/network_ues.json",
                                    "sdn_controller_ip": "127.0.0.1", "sdn_controller_port": 8080
                                    })

        yield

        testing.tearDown()

    @pytest.mark.parametrize("body, expected_error, error_msg", [
        (None, "Configuration must be a JSON object.", "A bad request error must have been raised in case of missing request body."),
        ('{"from": 12341412, "to": 1234897}', "Invalid JSON query document.", "A bad request error must have been raised in case of invalid request body."),
        ('{"service_function_chain": "sfc", "service_function_chain_instance": "sfci", "from": 12341412, "to": 1234897}',
         "Invalid JSON query document.", "A bad request error must have been raised in case of invalid request body."),
        ('{"service_function_chain": "sfc", "service_function_chain_instance": "sfc_1", "service_functions": {"invalid_json"}}',
         "Configuration must be a JSON object.", "A bad request error must have been raised in case of invalid JSON body."),
        ('{"from": 1528386860, "to":1528389860, "service_function_chain": "sfc", "service_function_chain_instance": "sfc_1", "service_functions": ["nginx", "minio"]}',
         "The service functions description should be represented with a JSON object.", "A bad request error must have been raised in case of invalid request body."),
        ('{"from": 1528386860, "to":1528389860, "service_function_chain_instance": "sfc_1", "service_functions": {"nginx": {"measurement_name": "nginx", "response_time_field": "mean(avg_processing_time)", "request_size_field": "mean(avg_request_size)", "response_size_field": "mean(avg_response_size"}}}',
         "Invalid JSON query document.", "A bad request error must have been raised in case of missing service function chain value in the request body"),
        ('{"from": 1528386860, "to":1628389860, "service_function_chain": "sfc", "service_functions": {"nginx": {"measurement_name": "nginx", "response_time_field": "mean(avg_processing_time)", "request_size_field": "mean(avg_request_size)", "response_size_field": "mean(avg_response_size"}}}',
         "Invalid JSON query document.", "A bad request error must have been raised in case of missing sfci ID in the request body"),
        ('{"from": "invalid", "to": "invalid", "service_function_chain": "sfc", "service_function_chain_instance": "sfc_1", "service_functions": {"nginx": {"measurement_name": "nginx", "response_time_field": "mean(avg_processing_time)", "request_size_field": "mean(avg_request_size)", "response_size_field": "mean(avg_response_size"}}}',
         "'from' parameter must be a timestamp integer", "A bad request error must have been raised in case of invalid timestamp parameters."),
        ('{"from": 123, "to": "invalid", "service_function_chain": "sfc", "service_function_chain_instance": "sfc_1", "service_functions": {"nginx": {"measurement_name": "nginx", "response_time_field": "mean(avg_processing_time)", "request_size_field": "mean(avg_request_size)", "response_size_field": "mean(avg_response_size"}}}',
        "'to' parameter must be a timestamp integer", "A bad request error must have been raised in case of invalid timestamp parameters."),
        ('{"from": -10, "to": 20, "service_function_chain": "sfc", "service_function_chain_instance": "sfc_1", "service_functions": {"nginx": {"measurement_name": "nginx", "response_time_field": "mean(avg_processing_time)", "request_size_field": "mean(avg_request_size)", "response_size_field": "mean(avg_response_size"}}}',
         "'from' parameter must be a positive timestamp integer or 0", "A bad request error must have been raised in case of negative timestamp parameters."),
        ('{"from": 20, "to": 10, "service_function_chain": "sfc", "service_function_chain_instance": "sfc_1", "service_functions": {"nginx": {"measurement_name": "nginx", "response_time_field": "mean(avg_processing_time)", "request_size_field": "mean(avg_request_size)", "response_size_field": "mean(avg_response_size"}}}',
         "'to' parameter timestamp must be greater than 'from' parameter timestamp", "A bad request error must have been raised in case of negative difference between 'to' and 'from' timestamp parameters."),
        ('{"to": 154342324, "service_function_chain": "sfc", "service_function_chain_instance": "sfc_1", "service_functions": {"nginx": {"measurement_name": "nginx", "response_time_field": "mean(avg_processing_time)", "request_size_field": "mean(avg_request_size)", "response_size_field": "mean(avg_response_size"}}}',
         "Invalid JSON query document.", "A bad request error must have been raised in case of missing 'from' timestamp parameters."),
        ('{"from": 21321232, "service_function_chain": "sfc", "service_function_chain_instance": "sfc_1", "service_functions": {"nginx": {"measurement_name": "nginx", "response_time_field": "mean(avg_processing_time)", "request_size_field": "mean(avg_request_size)", "response_size_field": "mean(avg_response_size"}}}',
         "Invalid JSON query document.", "A bad request error must have been raised in case of missing 'to' timestamp parameters."),
        ('{"from": 1528386860, "to":1628389860, "service_function_chain": "sfc", "service_function_chain_instance": "sfc_1", "service_functions": {"nginx": {"measurement_name": "nginx", "response_time_field": "mean(avg_processing_time)", "request_size_field": "mean(avg_request_size)", "response_size_field": "mean(avg_response_size"}}}',
         "Database for service function chain sfc not found.", "A bad request error must have been raised in case of a non-existing database."),
    ])
    def test_build_error_handling(self, body, expected_error, error_msg):
        """
        Tests the error handling of the graph build API endpoint by passing erroneous input and confirming an HTTPBadRequest was returned.

        :param body: body of the request to test
        :param expected_error: the expected error to be returned in the BadRequest response
        :param error_msg: the error message to pass in case of an error not being properly handled by the API endpoint (in other words, a test failure)
        """

        request = testing.DummyRequest()
        if body is not None:
            request.body = body
        request.body = request.body.encode(request.charset)

        error_raised = False
        try:
            GraphAPI(request).build_temporal_graph()
        except HTTPBadRequest as e:
            print(e)
            assert expected_error in str(e)
            error_raised = True
        assert error_raised, error_msg

    @patch('clmcservice.graphapi.views.uuid4')
    def test_build(self, uuid_mock, db_testing_data, graph_network_topology):
        """
        Tests the graph build API endpoint - it makes 2 API calls and checks that the expected graph was created (the influx data that's being used is reported to InfluxDB in the conftest file)

        :param uuid_mock: mock object to mock the behaviour of the uuid4 function
        :param db_testing_data: pair of time stamps - the from-to range of the generated influx test data and the graph db client object (this is a fixture from conftest)
        :param graph_network_topology: fixture to generate the network topology and clean it afterwards
        """

        from_timestamp, to_timestamp, graph_db = db_testing_data

        ue_nodes = set([node["name"] for node in graph_db.nodes.match("UserEquipment")])
        assert ue_nodes == set("ue" + str(i) for i in [2, 3, 6]), "UE nodes must have been created by the db_testing_data fixture"

        dc_nodes = set([node["name"] for node in graph_db.nodes.match("Cluster")])
        assert dc_nodes == set("DC" + str(i) for i in range(1, 7)), "Compute nodes must have been created by the db_testing_data fixture"

        switch_nodes = set([node["name"] for node in graph_db.nodes.match("Switch")])
        assert switch_nodes == set("127.0.0." + str(i) for i in range(1, 7)), "Switch nodes must have been created by the db_testing_data fixture"

        # Create a valid build request and send it to the API endpoint
        uuid_mock.return_value = "graph_test_build_uuid1"
        responses = graph_generator(from_timestamp, to_timestamp)  # generates 2 graphs by sending request to the build API and yielding back the responses for each request
        response = next(responses)

        graph_subresponse = response.pop("graph")

        assert response == {"database": "test_sfc"}, "Response must contain the database name"

        assert graph_subresponse["uuid"] == uuid_mock.return_value, "Request UUID must be attached to the response."
        assert graph_subresponse["time_range"]["from"] == from_timestamp * 10**9  # timestamp returned in nanoseconds
        assert graph_subresponse["time_range"]["to"] == to_timestamp * 10**9  # timestamp returned in nanoseconds
        assert set(graph_subresponse["endpoints"]) == {"minio_1_ep1", "nginx_1_ep1", "nginx_1_ep2"}, "Wrong list of new endpoints was returned by the build request"
        request_id = graph_subresponse["uuid"]

        # check that the appropriate nodes have been created
        sfp_names = set([node["name"] for node in graph_db.nodes.match("ServiceFunctionPackage")])
        assert sfp_names == {"nginx", "minio"}, "The graph must contain 2 service function packages - nginx and minio"
        sf_names = set([node["name"] for node in graph_db.nodes.match("ServiceFunction")])
        assert sf_names == {"nginx_1", "minio_1"}, "The graph must contain 2 service functions - nginx_1 and minio_1"
        endpoints = set([node["name"] for node in graph_db.nodes.match("Endpoint", uuid=request_id)])
        assert endpoints == {"minio_1_ep1", "nginx_1_ep1", "nginx_1_ep2"}, "The graph must contain 3 endpoints - minio_1_ep1, nginx_1_ep1, nginx_1_ep2"
        sfci_names = set([node["name"] for node in graph_db.nodes.match("ServiceFunctionChainInstance")])
        assert sfci_names == {"test_sfc_premium"}, "The graph must contain 1 service function chain instance - test_sfc_premium"
        sfc_names = set([node["name"] for node in graph_db.nodes.match("ServiceFunctionChain")])
        assert sfc_names == {"test_sfc"}, "The graph must contain 1 service function chain - test_sfc"

        reference_node = graph_db.nodes.match("Reference", uuid=request_id, sfci="test_sfc_premium", sfc="test_sfc").first()
        assert reference_node is not None and reference_node["from"] == from_timestamp * 10**9 and reference_node["to"] == to_timestamp * 10**9, "Reference node must have been created"

        # check the appropriate edges have been created
        self.check_exist_relationship(
            (
                ("minio_1_ep1", "Endpoint", "DC4", "Cluster", "hostedBy"),
                ("nginx_1_ep1", "Endpoint", "DC4", "Cluster", "hostedBy"),
                ("nginx_1_ep2", "Endpoint", "DC6", "Cluster", "hostedBy"),
                ("minio_1", "ServiceFunction", "minio_1_ep1", "Endpoint", "realisedBy"),
                ("nginx_1", "ServiceFunction", "nginx_1_ep1", "Endpoint", "realisedBy"),
                ("nginx_1", "ServiceFunction", "nginx_1_ep2", "Endpoint", "realisedBy"),
                ("minio_1", "ServiceFunction", "minio", "ServiceFunctionPackage", "instanceOf"),
                ("nginx_1", "ServiceFunction", "test_sfc_premium", "ServiceFunctionChainInstance", "utilizedBy"),
                ("minio_1", "ServiceFunction", "test_sfc_premium", "ServiceFunctionChainInstance", "utilizedBy"),
                ("nginx", "ServiceFunctionPackage", "test_sfc", "ServiceFunctionChain", "utilizedBy"),
                ("minio", "ServiceFunctionPackage", "test_sfc", "ServiceFunctionChain", "utilizedBy"),
                ("test_sfc_premium", "ServiceFunctionChainInstance", "test_sfc", "ServiceFunctionChain", "instanceOf"),
            ), graph_db, request_id
        )

        # check endpoint nodes have the correct properties
        for endpoint, response_time, request_size, response_size in (("minio_1_ep1", 9, 5760, 2033), ("nginx_1_ep1", 18.2, 2260, 9660), ("nginx_1_ep2", 22.2, 35600, 6420)):
            endpoint_node = graph_db.nodes.match("Endpoint", name=endpoint, uuid=request_id).first()
            assert endpoint_node["response_time"] == response_time, "Wrong response time property of endpoint node"
            # approximation is used to avoid long float numbers retrieved from influx, the test case ensures the results are different enough so that approximation of +-1 is good enough for testing
            assert endpoint_node["request_size"] == pytest.approx(request_size, 1), "Wrong request size attribute of endpoint node"
            assert endpoint_node["response_size"] == pytest.approx(response_size, 1), "Wrong response size attribute of endpoint node"

        # send a new request for a new service function chain instance and check the new subgraph has been created
        uuid_mock.return_value = "graph_test_build_uuid2"
        response = next(responses)

        graph_subresponse = response.pop("graph")

        assert response == {"database": "test_sfc"}, "Response must contain the database name"

        assert graph_subresponse["uuid"] == uuid_mock.return_value, "Request UUID must be attached to the response."
        assert graph_subresponse["time_range"]["from"] == from_timestamp * 10**9  # timestamp returned in nanoseconds
        assert graph_subresponse["time_range"]["to"] == to_timestamp * 10**9  # timestamp returned in nanoseconds
        assert set(graph_subresponse["endpoints"]) == {"minio_2_ep1", "apache_1_ep1"}, "Wrong list of new endpoints was returned by the build request"
        request_id = graph_subresponse["uuid"]

        # check the new nodes have been created
        assert graph_db.nodes.match("ServiceFunctionPackage", name="apache").first() is not None, "Service function package apache must have been added to the graph"

        for sf in ("apache_1", "minio_2"):
            assert graph_db.nodes.match("ServiceFunction", name=sf).first() is not None, "Service function {0} must have been added to the graph".format(sf)

        for ep in ("minio_2_ep1", "apache_1_ep1"):
            assert graph_db.nodes.match("Endpoint", name=ep, uuid=request_id).first() is not None, "Endpoint {0} must have been added to the graph".format(ep)

        assert graph_db.nodes.match("ServiceFunctionChainInstance", name="test_sfc_non_premium").first() is not None, "Service function chain instance test_sfc_non_premium must have been added to the graph"
        assert graph_db.nodes.match("ServiceFunctionChain", name="test_sfc").first() is not None, "Service function chain test_sfc must have been added to the graph"

        reference_node = graph_db.nodes.match("Reference", uuid=request_id, sfci="test_sfc_non_premium", sfc="test_sfc").first()
        assert reference_node is not None and reference_node["from"] == from_timestamp * 10**9 and reference_node["to"] == to_timestamp * 10**9, "Reference node must have been created"

        # check the appropriate edges have been created
        self.check_exist_relationship(
            (
                ("minio_2_ep1", "Endpoint", "DC5", "Cluster", "hostedBy"),
                ("apache_1_ep1", "Endpoint", "DC5", "Cluster", "hostedBy"),
                ("minio_2", "ServiceFunction", "minio_2_ep1", "Endpoint", "realisedBy"),
                ("apache_1", "ServiceFunction", "apache_1_ep1", "Endpoint", "realisedBy"),
                ("minio_2", "ServiceFunction", "minio", "ServiceFunctionPackage", "instanceOf"),
                ("apache_1", "ServiceFunction", "apache", "ServiceFunctionPackage", "instanceOf"),
                ("minio_2", "ServiceFunction", "test_sfc_non_premium", "ServiceFunctionChainInstance", "utilizedBy"),
                ("apache_1", "ServiceFunction", "test_sfc_non_premium", "ServiceFunctionChainInstance", "utilizedBy"),
                ("minio", "ServiceFunctionPackage", "test_sfc", "ServiceFunctionChain", "utilizedBy"),
                ("apache", "ServiceFunctionPackage", "test_sfc", "ServiceFunctionChain", "utilizedBy"),
                ("test_sfc_non_premium", "ServiceFunctionChainInstance", "test_sfc", "ServiceFunctionChain", "instanceOf")
            ), graph_db, request_id
        )

        # check endpoint nodes have the correct properties
        for endpoint, response_time, request_size, response_size in (("minio_2_ep1", 7, 2998, 3610), ("apache_1_ep1", 17.6, 1480, 7860)):
            endpoint_node = graph_db.nodes.match("Endpoint", name=endpoint, uuid=request_id).first()
            assert endpoint_node["response_time"] == response_time, "Wrong response time property of endpoint node"
            # approximation is used to avoid long float numbers retrieved from influx, the test case ensures the results are different enough so that approximation of +-1 is good enough for testing
            assert endpoint_node["request_size"] == pytest.approx(request_size, 1), "Wrong request size attribute of endpoint node"
            assert endpoint_node["response_size"] == pytest.approx(response_size, 1), "Wrong response size attribute of endpoint node"

        request = testing.DummyRequest()
        request.matchdict["sfc_id"] = "test_sfc"
        response = GraphAPI(request).delete_full_graph()
        assert response == {"deleted": 17}

    @patch('clmcservice.graphapi.views.uuid4')
    def test_delete_temporal(self, uuid_mock, db_testing_data, graph_network_topology):
        """
        Tests the delete API endpoint of the Graph API - the test depends on the build test to have been passed successfully so that graph_1_id and graph_2_id have been set

        :param uuid_mock: the mock object used to mimic the behaviour of the uuid.uuid4 function
        :param db_testing_data: pair of time stamps - the from-to range of the generated influx test data and the graph db client object (this is a fixture from conftest)
        :param graph_network_topology: fixture to generate the network topology and clean it afterwards
        """

        from_timestamp, to_timestamp, graph_db = db_testing_data

        # build test graphs
        responses = graph_generator(from_timestamp, to_timestamp)  # generates 2 graphs by sending request to the build API and yielding back the responses for each request

        uuid_mock.return_value = "graph_test_delete_uuid1"
        graph_1_test_id = uuid_mock.return_value
        next(responses)  # return value is the response, but we don't need it for this test

        uuid_mock.return_value = "graph_test_delete_uuid2"
        graph_2_test_id = uuid_mock.return_value
        next(responses)  # return value is the response, but we don't need it for this test

        # test erroneous behaviour
        request = testing.DummyRequest()
        request.matchdict["graph_id"] = "invalid_graph_id"
        error_raised = False
        try:
            GraphAPI(request).delete_temporal_graph()
        except HTTPNotFound:
            error_raised = True
        assert error_raised, "HTTP Not Found error must be raised in case of unrecognized subgraph ID"

        assert len(graph_db.nodes.match(uuid=graph_1_test_id)) == 4, "Graph build generator function is broken"
        assert len(graph_db.nodes.match(uuid=graph_2_test_id)) == 3, "Graph build generator function is broken"

        # delete the graph associated with graph_1_id
        request = testing.DummyRequest()
        request.matchdict["graph_id"] = graph_1_test_id
        response = GraphAPI(request).delete_temporal_graph()
        assert response == {"deleted": 4}, "Incorrect response when deleting temporal graph"
        assert len(graph_db.nodes.match(uuid=graph_1_test_id)) == 0, "Delete request is broken"

        # delete the graph associated with graph_2_id
        request = testing.DummyRequest()
        request.matchdict["graph_id"] = graph_2_test_id
        response = GraphAPI(request).delete_temporal_graph()
        assert response == {"deleted": 3}, "Incorrect response when deleting temporal graph"
        assert len(graph_db.nodes.match(uuid=graph_2_test_id)) == 0, "Delete request is broken"

        assert set([node["name"] for node in graph_db.nodes.match("Cluster")]) == set(["DC" + str(i) for i in range(1, 7)]), "Cluster nodes must not be deleted"
        assert set([node["name"] for node in graph_db.nodes.match("Switch")]) == set(["127.0.0." + str(i) for i in range(1, 7)]), "Switch nodes must not be deleted"
        assert set([node["name"] for node in graph_db.nodes.match("UserEquipment")]) == set(["ue" + str(i) for i in (2, 3, 6)]), "UE nodes must not be deleted"
        assert set([node["name"] for node in graph_db.nodes.match("ServiceFunctionEndpoint")]) == set(), "Endpoint nodes must have been deleted"
        assert set([node["name"] for node in graph_db.nodes.match("ServiceFunction")]) == {"nginx_1", "apache_1", "minio_1", "minio_2"}, "Service functions must not be deleted."
        assert set([node["name"] for node in graph_db.nodes.match("ServiceFunctionPackage")]) == {"nginx", "minio", "apache"}, "Service function packages must not be deleted"
        assert set([node["name"] for node in graph_db.nodes.match("ServiceFunctionChainInstance")]) == {"test_sfc_premium", "test_sfc_non_premium"}, "Service function chain instances must not be deleted"
        assert set([node["name"] for node in graph_db.nodes.match("ServiceFunctionChain")]) == {"test_sfc"}, "Service function chains must not be deleted"

        request = testing.DummyRequest()
        request.matchdict["sfc_id"] = "test_sfc"
        response = GraphAPI(request).delete_full_graph()
        assert response == {"deleted": 10}

    def test_delete_full(self, db_testing_data):
        """
        Tests the functionality to delete the full media service graph starting from a service function chain node down to an endpoint node.

        :param db_testing_data: pair of time stamps - the from-to range of the generated influx test data and the graph db client object (this is a fixture from conftest)

        :return:
        """

        from_timestamp, to_timestamp, graph_db = db_testing_data

        responses = graph_generator(from_timestamp, to_timestamp)  # generates 2 graphs by sending request to the build API and yielding back the responses for each request
        next(responses)
        next(responses)

        request = testing.DummyRequest()
        request.matchdict["sfc_id"] = "invalid_test_sfc"
        error_raised = False
        try:
            GraphAPI(request).delete_full_graph()
        except HTTPNotFound:
            error_raised = True
        assert error_raised, "Error must have been raised for invalid SFC identifier"

        request = testing.DummyRequest()
        request.matchdict["sfc_id"] = "test_sfc"
        response = GraphAPI(request).delete_full_graph()
        assert response == {"deleted": 17}

    @pytest.mark.parametrize("graph_id, endpoint, startpoint, error_type, error_msg", [
        ('e8cd4768-47dd-48cd-9c74-7f8926ddbad8', None, None, HTTPBadRequest, "HTTP Bad Request must be thrown in case of missing or invalid url parameters"),
        ('e8cd4768-47dd-48cd-9c74-7f8926ddbad8', None, "nginx", HTTPBadRequest, "HTTP Bad Request must be thrown in case of missing or invalid url parameters"),
        ('e8cd4768-47dd-48cd-9c74-7f8926ddbad8', "nginx_1_ep1", None, HTTPBadRequest, "HTTP Bad Request must be thrown in case of missing or invalid url parameters"),
        ('random-uuid', "nginx_1_ep1", "nginx", HTTPNotFound, "HTTP Not Found error must be thrown for an endpoint node with incorrect request ID"),
        ('random-uuid', "minio_1_ep1", "minio", HTTPNotFound, "HTTP Not Found error must be thrown for an endpoint node with incorrect request ID"),
    ])
    def test_rtt_error_handling(self, graph_id, endpoint, startpoint, error_type, error_msg):
        """
        Tests the error handling of the graph round trip time API endpoint - achieved by sending erroneous input in the request and verifying the appropriate error type has been returned.

        :param graph_id: the UUID of the subgraph
        :param endpoint: endpoint ID
        :param startpoint: the start node ID
        :param error_type: error type to expect as a response
        :param error_msg: error message in case of a test failure
        """

        request = testing.DummyRequest()
        request.matchdict["graph_id"] = graph_id
        if endpoint is not None:
            request.params["endpoint"] = endpoint
        if startpoint is not None:
            request.params["startpoint"] = startpoint
        error_raised = False
        try:
            GraphAPI(request).run_rtt_query()
        except error_type:
            error_raised = True
        assert error_raised, error_msg

    def test_rtt(self, db_testing_data, graph_network_topology):
        """
        Tests the rtt API endpoint of the Graph API.

        :param db_testing_data: pair of time stamps - the from-to range of the generated influx test data and the graph db client object (this is a fixture from conftest)
        :param graph_network_topology: fixture to generate the network topology and clean it afterwards
        """

        from_timestamp, to_timestamp, graph_db = db_testing_data

        # create a graph to use for RTT test by using the build API endpoint
        responses = graph_generator(from_timestamp, to_timestamp)  # generates 2 graphs by sending request to the build API and yielding back the responses for each request
        response = next(responses)
        request_id = response["graph"]["uuid"]

        # test more error case handling of the RTT API endpoint
        request = testing.DummyRequest()
        request.matchdict["graph_id"] = request_id
        request.params["endpoint"] = "nginx_1_ep1"
        request.params["compute"] = "DC1"
        error_raised = False
        try:
            GraphAPI(request).run_rtt_query()
        except HTTPBadRequest:
            error_raised = True
        assert error_raised, "HTTP Bad Request must be thrown in case of missing or invalid url parameters"

        request = testing.DummyRequest()
        request.matchdict["graph_id"] = request_id
        request.params["endpoint"] = "nginx_1_ep1"
        request.params["startpoint"] = "DC0"
        error_raised = False
        try:
            GraphAPI(request).run_rtt_query()
        except HTTPNotFound:
            error_raised = True
        assert error_raised, "HTTP Not Found error must be thrown for non existing compute node"

        request = testing.DummyRequest()
        request.matchdict["graph_id"] = request_id
        request.params["endpoint"] = "apache_1_ep1"
        request.params["startpoint"] = "DC1"
        error_raised = False
        try:
            GraphAPI(request).run_rtt_query()
        except HTTPNotFound:
            error_raised = True
        assert error_raised, "HTTP Not Found error must be thrown for a non existing endpoint"

        # test valid requests
        # go through the set of input/output (expected) parameters and assert actual results match with expected ones
        for startpoint, endpoint, forward_latencies, reverse_latencies, response_time, request_size, response_size, rtt, global_tags in (
            ("DC6", "nginx_1_ep2", [], [], 22.2, 35600, 6420, 22.2, {"flame_location": "DC6", "flame_sfe": "nginx_1_ep2", "flame_server": "DC6", "flame_sfc": "test_sfc", "flame_sfci": "test_sfc_premium", "flame_sfp": "nginx", "flame_sf": "nginx_1"}),
            ("127.0.0.6", "nginx_1_ep2", [0], [0], 22.2, 35600, 6420, 22.2, {"flame_location": "DC6", "flame_sfe": "nginx_1_ep2", "flame_server": "DC6", "flame_sfc": "test_sfc", "flame_sfci": "test_sfc_premium", "flame_sfp": "nginx", "flame_sf": "nginx_1"}),
            ("ue6", "nginx_1_ep2", [0, 0], [0, 0], 22.2, 35600, 6420, 22.2, {"flame_location": "DC6", "flame_sfe": "nginx_1_ep2", "flame_server": "DC6", "flame_sfc": "test_sfc", "flame_sfci": "test_sfc_premium", "flame_sfp": "nginx", "flame_sf": "nginx_1"}),
            ("DC2", "nginx_1_ep2", [0, 7.5, 15, 4.5, 0], [0, 4.5, 15, 7.5, 0], 22.2, 35600, 6420, 76.2, {"flame_location": "DC6", "flame_sfe": "nginx_1_ep2", "flame_server": "DC6", "flame_sfc": "test_sfc", "flame_sfci": "test_sfc_premium", "flame_sfp": "nginx", "flame_sf": "nginx_1"}),
            ("127.0.0.2", "nginx_1_ep2", [7.5, 15, 4.5, 0], [0, 4.5, 15, 7.5], 22.2, 35600, 6420, 76.2, {"flame_location": "DC6", "flame_sfe": "nginx_1_ep2", "flame_server": "DC6", "flame_sfc": "test_sfc", "flame_sfci": "test_sfc_premium", "flame_sfp": "nginx", "flame_sf": "nginx_1"}),
            ("DC3", "nginx_1_ep1", [0, 12.5, 0], [0, 12.5, 0], 18.2, 2260, 9660, 43.2, {"flame_location": "DC4", "flame_sfe": "nginx_1_ep1", "flame_server": "DC4", "flame_sfc": "test_sfc", "flame_sfci": "test_sfc_premium", "flame_sfp": "nginx", "flame_sf": "nginx_1"}),
            ("127.0.0.3", "nginx_1_ep1", [12.5, 0], [0, 12.5], 18.2, 2260, 9660, 43.2, {"flame_location": "DC4", "flame_sfe": "nginx_1_ep1", "flame_server": "DC4", "flame_sfc": "test_sfc", "flame_sfci": "test_sfc_premium", "flame_sfp": "nginx", "flame_sf": "nginx_1"}),
            ("ue3", "nginx_1_ep1", [0, 12.5, 0], [0, 12.5, 0], 18.2, 2260, 9660, 43.2, {"flame_location": "DC4", "flame_sfe": "nginx_1_ep1", "flame_server": "DC4", "flame_sfc": "test_sfc", "flame_sfci": "test_sfc_premium", "flame_sfp": "nginx", "flame_sf": "nginx_1"})
        ):
            request = testing.DummyRequest()
            request.matchdict["graph_id"] = request_id
            request.params["endpoint"] = endpoint
            request.params["startpoint"] = startpoint
            response = GraphAPI(request).run_rtt_query()
            # approximation is used to avoid incorrect float numbers comparison
            assert response.pop("request_size") == pytest.approx(request_size, 0.01), "Incorrect request size in RTT response"
            assert response.pop("response_size") == pytest.approx(response_size, 0.01), "Incorrect response size in RTT response"
            assert response.pop("round_trip_time") == pytest.approx(rtt, 0.01), "Incorrect RTT response"
            assert response == {"forward_latencies": forward_latencies, "reverse_latencies": reverse_latencies, "total_forward_latency": sum(forward_latencies), "total_reverse_latency": sum(reverse_latencies),
                                "response_time": response_time, "local_tags": {"traffic_source": startpoint}, "global_tags": global_tags}, "Incorrect RTT response"

        # send a new request for a new service function chain to create a second subgraph to test
        response = next(responses)
        request_id = response["graph"]["uuid"]

        # test valid requests
        # go through the set of input/output (expected) parameters and assert actual results match with expected ones
        for startpoint, endpoint, forward_latencies, reverse_latencies, response_time, request_size, response_size, rtt, global_tags in (
            ("DC5", "apache_1_ep1", [], [], 17.6, 1480, 7860, 17.6, {"flame_location": "DC5", "flame_sfe": "apache_1_ep1", "flame_server": "DC5", "flame_sfc": "test_sfc", "flame_sfci": "test_sfc_non_premium", "flame_sfp": "apache", "flame_sf": "apache_1"}),
            ("127.0.0.5", "apache_1_ep1", [0], [0], 17.6, 1480, 7860, 17.6, {"flame_location": "DC5", "flame_sfe": "apache_1_ep1", "flame_server": "DC5", "flame_sfc": "test_sfc", "flame_sfci": "test_sfc_non_premium", "flame_sfp": "apache", "flame_sf": "apache_1"}),
            ("DC5", "minio_2_ep1", [], [], 7, 2998.33, 3610, 7, {"flame_location": "DC5", "flame_sfe": "minio_2_ep1", "flame_server": "DC5", "flame_sfc": "test_sfc", "flame_sfci": "test_sfc_non_premium", "flame_sfp": "minio", "flame_sf": "minio_2"}),
            ("127.0.0.5", "minio_2_ep1", [0], [0], 7, 2998.33, 3610, 7, {"flame_location": "DC5", "flame_sfe": "minio_2_ep1", "flame_server": "DC5", "flame_sfc": "test_sfc", "flame_sfci": "test_sfc_non_premium", "flame_sfp": "minio", "flame_sf": "minio_2"}),
            ("DC3", "apache_1_ep1", [0, 9, 15, 0], [0, 15, 9, 0], 17.6, 1480, 7860, 65.6, {"flame_location": "DC5", "flame_sfe": "apache_1_ep1", "flame_server": "DC5", "flame_sfc": "test_sfc", "flame_sfci": "test_sfc_non_premium", "flame_sfp": "apache", "flame_sf": "apache_1"}),
            ("127.0.0.3", "apache_1_ep1", [9, 15, 0], [0, 15, 9], 17.6, 1480, 7860, 65.6, {"flame_location": "DC5", "flame_sfe": "apache_1_ep1", "flame_server": "DC5", "flame_sfc": "test_sfc", "flame_sfci": "test_sfc_non_premium", "flame_sfp": "apache", "flame_sf": "apache_1"}),
            ("ue3", "apache_1_ep1", [0, 9, 15, 0], [0, 15, 9, 0], 17.6, 1480, 7860, 65.6, {"flame_location": "DC5", "flame_sfe": "apache_1_ep1", "flame_server": "DC5", "flame_sfc": "test_sfc", "flame_sfci": "test_sfc_non_premium", "flame_sfp": "apache", "flame_sf": "apache_1"}),
            ("DC2", "minio_2_ep1", [0, 7.5, 15, 0], [0, 15, 7.5, 0], 7, 2998.33, 3610, 52, {"flame_location": "DC5", "flame_sfe": "minio_2_ep1", "flame_server": "DC5", "flame_sfc": "test_sfc", "flame_sfci": "test_sfc_non_premium", "flame_sfp": "minio", "flame_sf": "minio_2"}),
            ("127.0.0.2", "minio_2_ep1", [7.5, 15, 0], [0, 15, 7.5], 7, 2998.33, 3610, 52, {"flame_location": "DC5", "flame_sfe": "minio_2_ep1", "flame_server": "DC5", "flame_sfc": "test_sfc", "flame_sfci": "test_sfc_non_premium", "flame_sfp": "minio", "flame_sf": "minio_2"}),
            ("ue2", "minio_2_ep1", [0, 7.5, 15, 0], [0, 15, 7.5, 0], 7, 2998.33, 3610, 52, {"flame_location": "DC5", "flame_sfe": "minio_2_ep1", "flame_server": "DC5", "flame_sfc": "test_sfc", "flame_sfci": "test_sfc_non_premium", "flame_sfp": "minio", "flame_sf": "minio_2"})
        ):
            request = testing.DummyRequest()
            request.matchdict["graph_id"] = request_id
            request.params["endpoint"] = endpoint
            request.params["startpoint"] = startpoint
            response = GraphAPI(request).run_rtt_query()
            # approximation is used to avoid incorrect float numbers comparison
            assert response.pop("request_size") == pytest.approx(request_size, 0.01), "Incorrect request size in RTT response"
            assert response.pop("response_size") == pytest.approx(response_size, 0.01), "Incorrect response size in RTT response"
            assert response.pop("round_trip_time") == pytest.approx(rtt, 0.01), "Incorrect RTT response"
            assert response == {"forward_latencies": forward_latencies, "reverse_latencies": reverse_latencies, "total_forward_latency": sum(forward_latencies), "total_reverse_latency": sum(reverse_latencies),
                                "response_time": response_time, "local_tags": {"traffic_source": startpoint}, "global_tags": global_tags}, "Incorrect RTT response"

    @patch('clmcservice.graphapi.views.load')
    @patch('clmcservice.graphapi.views.open')
    @patch('clmcservice.graphapi.views.get')
    def test_build_network(self, http_get_mock, file_open_mock, json_load_mock, db_testing_data):
        """
        Tests the functionality to build the network graph.

        :param http_get_mock: mocks the HTTP GET function
        :param file_open_mock: mocks the open file function
        :param json_load_mock: mocks the JSON load function
        :param db_testing_data: fixture used to get a reference to the graph DB
        """

        from_timestamp, to_timestamp, graph_db = db_testing_data  # fixture, used to get reference to the graph DB

        # mock the responses from the sdn controller - 3 GET requests are executed, so we need 3 responses
        mock_response1 = Mock()
        mock_response1.status_code = 200
        mock_response1.json.return_value = sdn_switches
        mock_response2 = Mock()
        mock_response2.status_code = 200
        mock_response2.json.return_value = links
        mock_response3 = Mock()
        mock_response3.status_code = 200
        mock_response3.json.return_value = links
        # we are doing two calls to the API, hence need to repeat the responses
        http_get_mock.side_effect = [mock_response1, mock_response2, mock_response3] * 2

        # mock the behaviour of reading the clusters and ues mappping from files
        file_open_mock.return_value = MagicMock()  # use magic mock so that special methods (dunders) are auto generated
        # we are doing two calls to the API, hence need to repeat the results
        json_load_mock.side_effect = [clusters, ues] * 2

        assert set([node["name"] for node in graph_db.nodes.match("Cluster")]) == set(), "Cluster nodes must not be created before the build request"
        assert set([node["name"] for node in graph_db.nodes.match("Switch")]) == set(), "Switch nodes must not be created before the build request"
        assert set([node["name"] for node in graph_db.nodes.match("UserEquipment")]) == set(), "UE nodes must not be created before the build request"

        # sent request to build the network topology through a POST request
        request = testing.DummyRequest()
        response = GraphAPI(request).build_network_topology()
        assert response == {"new_switches_count": 6, "new_clusters_count": 6, "new_ues_count": 3}
        assert set([node["name"] for node in graph_db.nodes.match("Cluster")]) == set(["DC" + str(i) for i in range(1, 7)]), "Cluster nodes must have been created"
        assert set([node["name"] for node in graph_db.nodes.match("Switch")]) == set(["127.0.0." + str(i) for i in range(1, 7)]), "Switch nodes must have been created"
        assert set([node["name"] for node in graph_db.nodes.match("UserEquipment")]) == set(["ue" + str(i) for i in (2, 3, 6)]), "UE nodes must have been created"

        # send the same request and ensure that nothing else is created
        request = testing.DummyRequest()
        response = GraphAPI(request).build_network_topology()
        assert response == {"new_switches_count": 0, "new_clusters_count": 0, "new_ues_count": 0}
        assert set([node["name"] for node in graph_db.nodes.match("Cluster")]) == set(["DC" + str(i) for i in range(1, 7)])
        assert set([node["name"] for node in graph_db.nodes.match("Switch")]) == set(["127.0.0." + str(i) for i in range(1, 7)])
        assert set([node["name"] for node in graph_db.nodes.match("UserEquipment")]) == set(["ue" + str(i) for i in (2, 3, 6)])

        # clean up
        delete_network_graph(graph_db)

    @patch('clmcservice.graphapi.views.load')
    @patch('clmcservice.graphapi.views.open')
    @patch('clmcservice.graphapi.views.get')
    def test_build_and_update_network(self, http_get_mock, file_open_mock, json_load_mock, db_testing_data):
        """
        Tests the functionality to build and update the network graph.

        :param http_get_mock: mocks the HTTP GET function
        :param file_open_mock: mocks the open file function
        :param json_load_mock: mocks the JSON load function
        :param db_testing_data: fixture used to get a reference to the graph DB
        """

        from_timestamp, to_timestamp, graph_db = db_testing_data  # fixture, used to get reference to the graph DB

        # mock the responses from the sdn controller - 3 GET requests are executed, so we need 3 responses
        mock_response1 = Mock()
        mock_response1.status_code = 200
        mock_response1.json.return_value = sdn_switches
        mock_response2 = Mock()
        mock_response2.status_code = 200
        mock_response2.json.return_value = links
        mock_response3 = Mock()
        mock_response3.status_code = 200
        mock_response3.json.return_value = links
        # we are doing two calls to the API, hence need to repeat the responses
        http_get_mock.side_effect = [mock_response1, mock_response2, mock_response3] * 2

        # mock the behaviour of reading the clusters and ues mappping from files
        file_open_mock.return_value = MagicMock()  # use magic mock so that special methods (dunders) are auto generated
        # we are doing two calls to the API, hence need to repeat the results
        json_load_mock.side_effect = [clusters, ues] * 2

        # sent request to build the network topology through a PUT request
        request = testing.DummyRequest()
        response = GraphAPI(request).build_and_update_network_topology()
        assert response == {"new_switches_count": 6, "new_clusters_count": 6, "new_ues_count": 3}
        assert set([node["name"] for node in graph_db.nodes.match("Cluster")]) == set(["DC" + str(i) for i in range(1, 7)]), "Cluster nodes must have been created"
        assert set([node["name"] for node in graph_db.nodes.match("Switch")]) == set(["127.0.0." + str(i) for i in range(1, 7)]), "Switch nodes must have been created"
        assert set([node["name"] for node in graph_db.nodes.match("UserEquipment")]) == set(["ue" + str(i) for i in (2, 3, 6)]), "UE nodes must have been created"

        # fetch the first network link as test link
        test_link = links[0]
        test_link_source = switches[test_link["src-switch"]]
        test_link_destination = switches[test_link["dst-switch"]]
        test_link_source_node = graph_db.nodes.match("Switch", name=test_link_source).first()
        test_link_destination_node = graph_db.nodes.match("Switch", name=test_link_destination).first()
        # assert the nodes exist
        assert test_link_source_node is not None
        assert test_link_destination_node is not None

        # confirm the latency value of the edge
        old_latency = test_link["latency"] / 1000  # convert to seconds
        test_link_edge = graph_db.relationships.match(nodes=(test_link_source_node, test_link_destination_node), r_type="linkedTo").first()
        # assert the edge exists
        assert test_link_edge is not None
        assert test_link_edge["latency"] == old_latency, "Edge was not created with the correct latency value"

        # update the test link in the SDN controller mock data with a new latency
        new_latency = 5 * old_latency
        test_link["latency"] = new_latency * 1000  # convert to milliseconds

        # send the same request again and ensure that the latency of the network edge is updated
        request = testing.DummyRequest()
        GraphAPI(request).build_and_update_network_topology()

        graph_db.pull(test_link_edge)
        assert test_link_edge["latency"] == new_latency, "Edge was not updated with the correct latency value"

        # clean up
        delete_network_graph(graph_db)

    def test_delete_network(self, graph_network_topology, db_testing_data):
        """
        Tests the delete network graph functionality.

        :param graph_network_topology: fixture to generate the network topology and clean it afterwards
        :param db_testing_data: fixture used to get a reference to the graph DB
        """

        from_timestamp, to_timestamp, graph_db = db_testing_data  # fixture, used to get reference to the graph DB

        assert set([node["name"] for node in graph_db.nodes.match("Cluster")]) == set(["DC" + str(i) for i in range(1, 7)]), "Cluster nodes must have been created before the delete request"
        assert set([node["name"] for node in graph_db.nodes.match("Switch")]) == set(["127.0.0." + str(i) for i in range(1, 7)]), "Switch nodes must have been created before the delete request"
        assert set([node["name"] for node in graph_db.nodes.match("UserEquipment")]) == set(["ue" + str(i) for i in (2, 3, 6)]), "UE nodes must have been created before the delete request"

        # send the first delete request
        request = testing.DummyRequest()
        response = GraphAPI(request).delete_network_topology()
        assert response == {"deleted_switches_count": 6, "deleted_clusters_count": 6, "deleted_ues_count": 3}

        assert set([node["name"] for node in graph_db.nodes.match("Cluster")]) == set(), "Cluster nodes must have been deleted"
        assert set([node["name"] for node in graph_db.nodes.match("Switch")]) == set(), "Switch nodes must have been deleted"
        assert set([node["name"] for node in graph_db.nodes.match("UserEquipment")]) == set(), "UE nodes must have been deleted"

        # send the same request and ensure that nothing else is deleted
        request = testing.DummyRequest()
        response = GraphAPI(request).delete_network_topology()
        assert response == {"deleted_switches_count": 0, "deleted_clusters_count": 0, "deleted_ues_count": 0}

        assert set([node["name"] for node in graph_db.nodes.match("Cluster")]) == set()
        assert set([node["name"] for node in graph_db.nodes.match("Switch")]) == set()
        assert set([node["name"] for node in graph_db.nodes.match("UserEquipment")]) == set()

    @pytest.mark.parametrize("body, expected_error, error_msg", [
        (None, "Configuration must be a JSON object.", "A bad request error must have been raised in case of missing request body."),
        ('{"query_period": 45, "results_measurement_name": "graph"}', "Invalid JSON query document.", "A bad request error must have been raised in case of invalid request body."),
        ('{"service_function_chain": "sfc", "service_function_chain_instance": "sfci", "query_period": 30, "results_measurement_name": "graph"}',
         "Invalid JSON query document.", "A bad request error must have been raised in case of invalid request body."),
        ('{"service_function_chain": "sfc", "service_function_chain_instance": "sfc_1", "service_functions": {"invalid_json"}}',
         "Configuration must be a JSON object.", "A bad request error must have been raised in case of invalid JSON body."),
        ('{"query_period": 45, "results_measurement_name": "graph", "service_function_chain": "sfc", "service_function_chain_instance": "sfc_1", "service_functions": ["nginx", "minio"]}',
         "The service functions description should be represented with a JSON object.", "A bad request error must have been raised in case of invalid request body."),
        ('{"query_period": 45, "results_measurement_name": "graph", "service_function_chain_instance": "sfc_1", "service_functions": {"nginx": {"measurement_name": "nginx", "response_time_field": "mean(avg_processing_time)", "request_size_field": "mean(avg_request_size)", "response_size_field": "mean(avg_response_size"}}}',
         "Invalid JSON query document.", "A bad request error must have been raised in case of missing service function chain value in the request body"),
        ('{"query_period": 45, "results_measurement_name": "graph", "service_function_chain": "sfc", "service_functions": {"nginx": {"measurement_name": "nginx", "response_time_field": "mean(avg_processing_time)", "request_size_field": "mean(avg_request_size)", "response_size_field": "mean(avg_response_size"}}}',
         "Invalid JSON query document.", "A bad request error must have been raised in case of missing sfci ID in the request body"),
        ('{"query_period": "invalid", "results_measurement_name": "graph", "service_function_chain": "sfc", "service_function_chain_instance": "sfc_1", "service_functions": {"nginx": {"measurement_name": "nginx", "response_time_field": "mean(avg_processing_time)", "request_size_field": "mean(avg_request_size)", "response_size_field": "mean(avg_response_size"}}}',
         "'query_period' parameter must be an integer", "A bad request error must have been raised in case of invalid query period parameter."),
        ('{"query_period": -30, "results_measurement_name": "graph", "service_function_chain": "sfc", "service_function_chain_instance": "sfc_1", "service_functions": {"nginx": {"measurement_name": "nginx", "response_time_field": "mean(avg_processing_time)", "request_size_field": "mean(avg_request_size)", "response_size_field": "mean(avg_response_size"}}}',
        "'query_period' parameter must be a positive integer.", "A bad request error must have been raised in case of non-positive query period parameter."),
        ('{"results_measurement_name": "graph", "service_function_chain": "sfc", "service_function_chain_instance": "sfc_1", "service_functions": {"nginx": {"measurement_name": "nginx", "response_time_field": "mean(avg_processing_time)", "request_size_field": "mean(avg_request_size)", "response_size_field": "mean(avg_response_size"}}}',
         "Invalid JSON query document.", "A bad request error must have been raised in case of missing 'query period' parameter."),
        ('{"query_period": 30, "service_function_chain": "sfc", "service_function_chain_instance": "sfc_1", "service_functions": {"nginx": {"measurement_name": "nginx", "response_time_field": "mean(avg_processing_time)", "request_size_field": "mean(avg_request_size)", "response_size_field": "mean(avg_response_size"}}}',
         "Invalid JSON query document.", "A bad request error must have been raised in case of missing 'results_measurement_name' parameter."),
        ('{"query_period": 45, "results_measurement_name": "graph", "service_function_chain": "sfc", "service_function_chain_instance": "sfc_1", "service_functions": {"nginx": {"measurement_name": "nginx", "response_time_field": "mean(avg_processing_time)", "request_size_field": "mean(avg_request_size)", "response_size_field": "mean(avg_response_size"}}}',
         "Database for service function chain sfc not found.", "A bad request error must have been raised in case of a non-existing database."),
    ])
    def test_pipeline_error_handling(self, body, expected_error, error_msg):
        """
        Tests the handling of invalid requests for the execute pipeline graph API endpoint.

        :param body: body of the request to test
        :param expected_error: the expected error to be returned in the BadRequest response
        :param error_msg: the error message to pass in case of an error not being properly handled by the API endpoint (in other words, a test failure)
        """

        request = testing.DummyRequest()
        if body is not None:
            request.body = body
        request.body = request.body.encode(request.charset)

        error_raised = False
        try:
            GraphAPI(request).execute_graph_pipeline()
        except HTTPBadRequest as e:
            print(e)
            assert expected_error in str(e)
            error_raised = True
        assert error_raised, error_msg

    @patch('clmcservice.graphapi.views.Graph')
    @patch('clmcservice.graphapi.views.Popen')
    @patch('clmcservice.graphapi.views.uuid4')
    def test_execute_graph_pipeline(self, uuid_mock, popen_mock, graph_mock):
        """
        Tests the functionality to start a pipeline script executing the graph API workflow - build, query, delete.

        :param uuid_mock: mock object for the uuid generator function
        :param popen_mock: mock object for the process creation function
        :param graph_mock: mock object for the graph DB client
        """

        # mock the behaviour of the uuid4 function
        uuid_mock.return_value = "monitor_test_uuid1"

        # mock the behaviour of the Popen class
        pid_property_mock = PropertyMock(return_value=111)
        returncode_property_mock = PropertyMock(return_value=None)
        popen_intance_mock = Mock()
        type(popen_intance_mock).pid = pid_property_mock  # a property mock cannot be attached directly to the mock object, hence use its type object
        type(popen_intance_mock).returncode = returncode_property_mock  # a property mock cannot be attached directly to the mock object, hence use its type object
        popen_mock.return_value = popen_intance_mock

        # mock the behaviour of the graph nodes match function
        nodes_matcher_mock = Mock(return_value=[{"name": "ue1"}, {"name": "ue2"}, {"name": "ue3"}])  # the API method expects node objects with 'name' key (node['name'])
        graph_mock.return_value.nodes.match = nodes_matcher_mock

        # check proper behaviour of the API endpoint
        service_functions = dict(nginx={"measurement_name": "nginx", "response_time_field": "mean(avg_processing_time)",
                                        "request_size_field": "mean(avg_request_size)", "response_size_field": "mean(avg_response_size)"},
                                 minio={"measurement_name": "minio_http", "response_time_field": "mean(total_processing_time)/mean(total_requests_count)",
                                        "request_size_field": "mean(total_requests_size)/mean(total_requests_count)", "response_size_field": "mean(total_response_size)/mean(total_requests_count)"})
        monitor_json_body = dict(service_function_chain="test_sfc", service_function_chain_instance="test_sfc_premium", service_functions=service_functions)
        monitor_json_body["query_period"] = 30
        monitor_json_body["results_measurement_name"] = "graph_measurements"
        body = dumps(monitor_json_body)
        request = testing.DummyRequest()
        request.body = body.encode(request.charset)
        response = GraphAPI(request).execute_graph_pipeline()
        assert response == {"uuid": uuid_mock.return_value, "database": "test_sfc"}

        # the API endpoint should have done this - appending the list of UEs to the JSON config of the pipeline script
        monitor_json_body["ues"] = [node["name"] for node in nodes_matcher_mock.return_value]

        # assert that the graph pipeline script is ran with the JSON config that was received in the request along with the UEs
        # we expect one call to the Popen() mock (first 0 index), we take the proper arguments, not kwargs (second 0 index)
        actual_call_arguments = popen_mock.call_args_list[0][0]
        # expecting one argument which is a list
        assert len(actual_call_arguments) == 1 and type(actual_call_arguments[0]) == list, "Incorrect call to Popen"
        assert actual_call_arguments[0][0] == "graph-pipeline.sh", "Incorrect graph pipeline script name"
        assert loads(actual_call_arguments[0][1]) == monitor_json_body, "Incorrect JSON configuration passed to pipeline script"
        pid_property_mock.assert_called_once_with()  # assert that the process ID attribute was called and saved
        returncode_property_mock.assert_called_once_with()  # assert that the process return code attribute was called to check if the process has started successfully
        nodes_matcher_mock.assert_called_once_with("UserEquipment")  # assert that the graph nodes match function has been called with "UserEquipment" as argument

        # check that the process ID was saved
        assert MonitoringProcess.exists(uuid_mock.return_value), "Request identifier was not saved during the request processing"
        assert MonitoringProcess.get(uuid_mock.return_value) == pid_property_mock.return_value, "Incorrect PID was saved during the request processing"

        # check erroneous behaviour - process returns a value immediately
        returncode_property_mock.return_value = -1
        request = testing.DummyRequest()
        request.body = body.encode(request.charset)
        error_raised = False
        try:
            GraphAPI(request).execute_graph_pipeline()
        except HTTPInternalServerError:
            error_raised = True
        assert error_raised, "Expecting a 500 HTTP error if the process terminated immediately after it was started"

        # assert that the graph pipeline script was still started with the JSON config that was received in the request along with the UEs
        # we expect a second call to the Popen() mock (index 1), we take the proper arguments, not kwargs (first 0 index)
        actual_call_arguments = popen_mock.call_args_list[1][0]
        # expecting one argument which is a list
        assert len(actual_call_arguments) == 1 and type(actual_call_arguments[0]) == list, "Incorrect call to Popen"
        assert actual_call_arguments[0][0] == "graph-pipeline.sh", "Incorrect graph pipeline script name"
        assert loads(actual_call_arguments[0][1]) == monitor_json_body, "Incorrect JSON configuration passed to pipeline script"
        pid_property_mock.assert_called_with()  # assert that the process ID attribute was called and saved
        returncode_property_mock.assert_called_with()  # assert that the process return code attribute was called to check if the process has started successfully
        nodes_matcher_mock.assert_called_with("UserEquipment")  # assert that the graph nodes match function has been called with "UserEquipment" as argument

        # check erroneous behaviour - no UE nodes are found, expecting bad request
        nodes_matcher_mock.return_value = []
        request = testing.DummyRequest()
        request.body = body.encode(request.charset)
        error_raised = False
        try:
            GraphAPI(request).execute_graph_pipeline()
        except HTTPBadRequest:
            error_raised = True
        assert error_raised, "Expecting a 400 HTTP Bad Request error if no UE nodes were found (network topology not built)."
        assert len(popen_mock.call_args_list) == 2, "No subprocess should be started if the UE nodes list is empty (network topology not built)"
        nodes_matcher_mock.assert_called_with("UserEquipment")  # assert that the graph nodes match function has been called with "UserEquipment" as argument

    @patch('clmcservice.graphapi.views.Process')
    def test_get_graph_pipeline_status(self, mock_process):
        """
        Tests the functionality to fetch the status of a graph monitoring script.

        :param mock_process: mock object to mimic the behavior of the psutil.Process functionality
        """

        # mock a monitoring process
        pid = 111
        reqid = "test_request_id"
        MonitoringProcess.add({"request_id": reqid, "process_id": pid})

        # test behaviour with not-existing request UUID
        request = testing.DummyRequest()
        request.matchdict["request_id"] = "unknown-request-uuid"
        error_raised = False
        try:
            GraphAPI(request).get_graph_pipeline_status()
        except HTTPNotFound:
            error_raised = True
        assert error_raised, "Error must have been raised for unrecognised request UUID."

        # test the behaviour when the PID doesn't exist or another OSError is thrown
        mock_process.side_effect = NoSuchProcess("error")
        request = testing.DummyRequest()
        request.matchdict["request_id"] = reqid
        response = GraphAPI(request).get_graph_pipeline_status()
        assert response == {"msg": "Monitoring process has been stopped or killed or terminated before this request was executed."}

        # test behaviour with existing request UUID and existing PID
        assert MonitoringProcess.exists(reqid)
        mock_process.side_effect = None
        mock_process.return_value.status = Mock(return_value=STATUS_SLEEPING)
        request = testing.DummyRequest()
        request.matchdict["request_id"] = reqid
        response = GraphAPI(request).get_graph_pipeline_status()
        assert response == {"status": STATUS_SLEEPING, "msg": "Successfully fetched status of graph pipeline process."}
        mock_process.return_value.status.assert_called_with()

    @patch('clmcservice.graphapi.views.Process')
    def test_stop_graph_pipeline(self, mock_process):
        """
        Tests the functionality to stop a graph monitoring script.

        :param mock_process: mock object to mimic the behavior of the psutil.Process functionality
        """

        # mock a monitoring process
        pid = 111
        reqid = "test_request_id"
        MonitoringProcess.add({"request_id": reqid, "process_id": pid})

        # test behaviour with not-existing request UUID
        request = testing.DummyRequest()
        request.matchdict["request_id"] = "unknown-request-uuid"
        error_raised = False
        try:
            GraphAPI(request).stop_graph_pipeline()
        except HTTPNotFound:
            error_raised = True
        assert error_raised, "Error must have been raised for unrecognised request UUID."

        # test the behaviour when the PID doesn't exist or another OSError is thrown
        mock_process.side_effect = NoSuchProcess("error")
        request = testing.DummyRequest()
        request.matchdict["request_id"] = reqid
        response = GraphAPI(request).stop_graph_pipeline()
        assert response == {"msg": "Monitoring process has been stopped or killed or terminated before this request was executed."}

        # test behaviour with existing request UUID and existing PID
        MonitoringProcess.add({"request_id": reqid, "process_id": pid})
        assert MonitoringProcess.exists(reqid)
        mock_process.side_effect = None
        mock_process.return_value = Mock()
        request = testing.DummyRequest()
        request.matchdict["request_id"] = reqid
        response = GraphAPI(request).stop_graph_pipeline()
        assert response == {"msg": "Monitoring process has been successfully stopped."}
        mock_process.return_value.terminate.assert_called_with()
        mock_process.return_value.wait.assert_called_with(timeout=3)
        assert not MonitoringProcess.exists(reqid), "Request ID must be removed when the process is killed."

    @staticmethod
    def check_exist_relationship(relationships_tuple, graph, uuid):
        """
        Iterates through a tuple of relationships and checks that each of those exists - a utility method to be reused for testing.

        :param relationships_tuple: the tuple to iterate
        :param graph: the graph object
        :param uuid: the uuid of the request
        """

        for relationship in relationships_tuple:
            from_node_name, from_node_type, to_node_name, to_node_type, relationship_type = relationship
            if from_node_type == "Endpoint":
                from_node = graph.nodes.match(from_node_type, name=from_node_name, uuid=uuid).first()
            else:
                from_node = graph.nodes.match(from_node_type, name=from_node_name).first()
            assert from_node is not None  # IMPORTANT, assert the from_node exists, otherwise the py2neo RelationshipMatcher object assumes you are looking for any node (instead of raising an error)

            if to_node_type == "Endpoint":
                to_node = graph.nodes.match(to_node_type, name=to_node_name, uuid=uuid).first()
            else:
                to_node = graph.nodes.match(to_node_type, name=to_node_name).first()
            assert to_node is not None  # IMPORTANT, assert the from_node exists, otherwise the py2neo RelationshipMatcher object assumes you are looking for any node (instead of raising an error)

            assert graph.relationships.match(nodes=(from_node, to_node), r_type=relationship_type).first() is not None, "Graph is missing a required relationship"


def graph_generator(from_timestamp, to_timestamp):
    """
    Utility function (generator) used to send 2 valid graph build requests to the graph API - yields the response of each.

    :param from_timestamp: 'from' timestamp
    :param to_timestamp: 'to' timestamp
    """

    service_functions = dict(nginx={"measurement_name": "nginx", "response_time_field": "mean(avg_processing_time)",
                                    "request_size_field": "mean(avg_request_size)", "response_size_field": "mean(avg_response_size)"},
                             minio={"measurement_name": "minio_http", "response_time_field": "mean(total_processing_time)/mean(total_requests_count)",
                                    "request_size_field": "mean(total_requests_size)/mean(total_requests_count)", "response_size_field": "mean(total_response_size)/mean(total_requests_count)"})
    build_json_body = dict(service_function_chain="test_sfc", service_function_chain_instance="test_sfc_premium", service_functions=service_functions)
    build_json_body["from"] = from_timestamp
    build_json_body["to"] = to_timestamp
    body = dumps(build_json_body)
    request = testing.DummyRequest()
    request.body = body.encode(request.charset)
    response = GraphAPI(request).build_temporal_graph()

    yield response

    service_functions = dict(minio={"measurement_name": "minio_http", "response_time_field": "mean(total_processing_time)/mean(total_requests_count)",
                                    "request_size_field": "mean(total_requests_size)/mean(total_requests_count)", "response_size_field": "mean(total_response_size)/mean(total_requests_count)"},
                             apache={"measurement_name": "apache", "response_time_field": "mean(avg_processing_time)",
                                     "request_size_field": "mean(avg_request_size)", "response_size_field": "mean(avg_response_size)"})
    build_json_body = dict(service_function_chain="test_sfc", service_function_chain_instance="test_sfc_non_premium", service_functions=service_functions)
    build_json_body["from"] = from_timestamp
    build_json_body["to"] = to_timestamp
    body = dumps(build_json_body)
    request = testing.DummyRequest()
    request.body = body.encode(request.charset)
    response = GraphAPI(request).build_temporal_graph()

    yield response
